/*********************************************************
 * *Copyright (C) 2015 Neil Horman
 * *This program is free software; you can redistribute it and\or modify
 * *it under the terms of the GNU General Public License as published 
 * *by the Free Software Foundation; either version 2 of the License,
 * *or  any later version.
 * *
 * *This program is distributed in the hope that it will be useful,
 * *but WITHOUT ANY WARRANTY; without even the implied warranty of
 * *MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
 * *GNU General Public License for more details.
 * *
 * *File: freight-db.c
 * *
 * *Author:Neil Horman
 * *
 * *Date: April 16, 2015
 * *
 * *Description API for DB connections
 * *********************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <alloca.h>
#include <string.h>
#include <freight-common.h>
#include <freight-log.h>
#include <freight-db.h>

struct db_api *api = NULL;

static const char* channel_map[] = {
	[CHAN_CONTAINERS] = "containers",
	[CHAN_TENNANT_HOSTS] = "tennant_hosts"
};

static char *tablenames[TABLE_MAX] = {
        [TABLE_TENNANTS] = "tennants",
        [TABLE_NODES] = "nodes",
        [TABLE_TENNANT_HOSTS] = "tennant_hosts",
        [TABLE_YUM_CONFIG] = "yum_config",
        [TABLE_CONTAINERS] = "containers",
	[TABLE_NETWORKS] = "networks",
	[TABLE_NETMAP] = "net_container_map",
	[TABLE_EVENTS] = "event_table"
};

/*
 * This table maps the human readable column names
 * to the numeric columns that each table returns.
 * The array indicies are:
 * TENNANT, HOSTNAME, STATE, NAME, URL, INAME, CNAME, PROXYPASS, TYPE, CONFIG
 * Note, that the net_contaienr_map uses INAME for the container name and CNAME for the network name
 */

static int db_col_map[TABLE_MAX][COL_MAX] = {
 [TABLE_TENNANTS] =		{ 0, -1, -1, -1, -1, -1, -1,  1, -1},
 [TABLE_NODES] =		{-1,  0,  1, -1, -1, -1, -1, -1, -1},
 [TABLE_TENNANT_HOSTS] = 	{ 1,  0, -1, -1, -1, -1, -1, -1, -1},
 [TABLE_YUM_CONFIG] =		{ 2, -1, -1,  0,  1, -1, -1, -1, -1},
 [TABLE_CONTAINERS] =		{ 0,  3,  4, -1, -1,  1,  2, -1, -1},
 [TABLE_NETWORKS] =		{ 1, -1,  2,  0, -1, -1, -1, -1,  3},
 [TABLE_NETMAP] =		{ 0, -1, -1, -1, -1,  1,  2, -1, -1}, 
 [TABLE_EVENTS] =		{ -1,-1, -1,  0, -1, -1, -1, -1,  1}
};


struct channel_callback {
	enum event_rc (*hndl)(const enum listen_channel chnl, const char *extra, const struct agent_config *acfg);
	enum listen_channel chnl;
	const struct agent_config *acfg;
	struct channel_callback *next;
};

static struct channel_callback *callbacks = NULL;

static int __chn_subscribe(const struct agent_config *acfg,
		    const char *lcmd,
		    const char *chnl)

{
	if (!api->subscribe)
		return -EOPNOTSUPP;
	return api->subscribe(lcmd, chnl, acfg);
}

int channel_add_tennant_subscription(const struct agent_config *acfg,
				     const enum listen_channel chn,
				     const char *tennant)
{
	char *chname;
	struct channel_callback *idx;


	/*
	 * need to ensure that we already have a subscription
	 */
	for (idx = callbacks; idx != NULL; idx = idx->next) {
		if (idx->chnl == chn)
			goto update;
	}
	return -ENOENT;

update:
	chname = strjoin("\"", channel_map[chn], "-", tennant, "\"", NULL);
	
	return __chn_subscribe(acfg, "LISTEN", chname);
}

int channel_del_tennant_subscription(const struct agent_config *acfg,
				     const enum listen_channel chn,
				     const char *tennant)
{
	char *chname;

	chname = strjoin("\"", channel_map[chn], "-", tennant, "\"", NULL);
	
	return __chn_subscribe(acfg, "UNLISTEN", chname);
}

int channel_subscribe(const struct agent_config *acfg,
		      const enum listen_channel chn,
		      enum event_rc (*hndl)(const enum listen_channel chnl, const char *extra, const struct agent_config *acfg))
{

	struct channel_callback *tmp;
	char *chname;
	int rc;

	tmp = callbacks;
	/*
	 * Look for callbacks already subscribed to this channel
	 */
	while (tmp) {
		if (tmp->chnl == chn)
			break;
		tmp = tmp->next;
	}

	if (tmp)
		return -EBUSY;

	tmp = calloc(1, sizeof(struct channel_callback));
	if (!tmp)
		return -ENOMEM;

	tmp->hndl = hndl;
	tmp->chnl = chn;
	tmp->acfg = acfg;
	tmp->next = callbacks;
	callbacks = tmp;


	/*
	 * We actually want to subscribe to n channels here, <name>-<hostname>
	 * channel, and the <name>-<tennant> channel, for each tennant that this
	 * node is subscribed to.
	 */
	chname = strjoina("\"", channel_map[chn],"-", acfg->cmdline.hostname, "\"", NULL);
	rc = __chn_subscribe(acfg, "LISTEN", chname);

	/*
	 * once we are subscribed, make a dummy call to the handler for an initial table scan
	 */
	if (!rc)
		hndl(chn, NULL, acfg);

	return rc;
}


void channel_unsubscribe(const struct agent_config *acfg,
			 const enum listen_channel chn)
{
	struct channel_callback *tmp, *prev;
	char *chname;

	tmp = prev = callbacks;

	while (tmp) {
		if (tmp->chnl == chn)
			break;
		prev = tmp;
		tmp = tmp->next;
	}

	if (!tmp)
		return;

	prev->next = tmp->next;

	chname = strjoina("\"", channel_map[chn],"-", acfg->cmdline.hostname, "\"", NULL);
	__chn_subscribe(acfg, "UNLISTEN", chname);

}

enum event_rc event_dispatch(const char *chn, const char *extra)
{
	struct channel_callback *tmp;

	tmp = callbacks;
	while (tmp) {
		if (!strncmp(channel_map[tmp->chnl], chn, strlen(channel_map[tmp->chnl])))
			return tmp->hndl(tmp->chnl, extra, tmp->acfg);
		tmp = tmp->next;
	}

	return EVENT_NOCHAN;
}


const char* get_tablename(enum db_table id)
{
	return tablenames[id];
}

const enum db_table get_tableid(const char *name)
{
	int i;

	for (i=0; i < TABLE_MAX; i++) {
		if (!strcmp(name, tablenames[i]))
			return i;
	}

	return TABLE_MAX;
}


struct tbl *alloc_tbl(int rows, int cols, enum db_table type)
{
	int r;

	struct tbl *table = calloc(1, sizeof(struct tbl));
	if (!table)
		goto out;
	table->rows = rows;
	table->cols = cols;
	table->type = type;

	table->value = calloc(1, sizeof(char **)*rows);
	if (!table->value)
		goto out_free_table;

	for(r = 0; r < rows; r++) {
		table->value[r] = calloc(1, sizeof(char *)*cols);
		if (!table->value[r])
			goto out_free_cols;
	}

	return table;

out_free_cols:
	for (r = 0; r < rows; r++ )
		free(table->value[r]);
	free(table->value);
out_free_table:
	free(table);
	table = NULL;
out:
	return table;
}

void free_tbl(struct tbl *table)
{
	int i, j;

	if (!table)
		return;

	for (i = 0; i < table->rows; i++)
		for (j = 0; j < table->cols; j++)
			free(table->value[i][j]);

	for (i = 0; i < table->rows; i++)
		free(table->value[i]);

	free(table->value);
	free(table);
}

int is_tbl_empty(struct tbl *table)
{
	return (table->rows == 0);
}

void * lookup_tbl(struct tbl *table, int row, enum table_col col)
{
	if ((long long)(table->value[row][db_col_map[table->type][col]]) == -1)
		return NULL;

	return table->value[row][db_col_map[table->type][col]];
}

char* get_tennant_proxy_pass(const char *user, const struct agent_config *acfg)
{
	struct tbl *table;
	char *filter;
	char *pass = NULL;

	if (!api->get_table)
		return NULL;

	filter = strjoina("tennant = '", user, "'", NULL);

	table = api->get_table(TABLE_TENNANTS, "*", filter, acfg);

	if (!table->rows)
		goto out;

	pass = lookup_tbl(table, 0, COL_PROXYPASS);

	if (!pass)
		goto out;

	pass = strdup(pass);
out:
	free_tbl(table);
	return pass;
}

int add_repo(const char *name, const char *url,
	     const char *tennant,
	     const struct agent_config *acfg)
{
	char *sql; 

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("INSERT INTO yum_config VALUES ('", name, "', '",
			url, "', '", tennant, "')", NULL);

	return api->send_raw_sql(sql, acfg);
}

extern int del_repo(const char *name,
		    const char *tennant,
		    const struct agent_config *acfg)
{
	char *sql;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("DELETE FROM yum_config WHERE name = '", name, 
		"' AND tennant='", tennant, "'", NULL);

	return api->send_raw_sql(sql, acfg);
}


int add_host(const char *hostname,
	     const struct agent_config *acfg)
{
	char *sql;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("INSERT INTO nodes VALUES ('", hostname, "', 'offline')", NULL);

	return api->send_raw_sql(sql, acfg);
}

int del_host(const char *hostname,
	     const struct agent_config *acfg)
{
	char *sql;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("DELETE FROM nodes WHERE hostname = '", hostname, "'", NULL);

	return api->send_raw_sql(sql, acfg);
}

int subscribe_host(const char *host,
		   const char *tennant,
		   const struct agent_config *acfg)
{
	char *sql;
	int rc;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("INSERT INTO tennant_hosts VALUES('", host, "', '", tennant, "')", NULL);

	rc = api->send_raw_sql(sql, acfg);
	if (rc)
		return rc;

	return notify_host(CHAN_TENNANT_HOSTS, host, acfg);
}

int unsubscribe_host(const char *tenant,
		     const char *host,
		     const struct agent_config *acfg)
{
	char *sql;
	int rc;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("DELETE FROM tennant_hosts WHERE hostname = '", 
			host,"' AND tennant = '", tenant, "'", NULL);

	rc = api->send_raw_sql(sql, acfg);
	if (rc)
		return rc;

	return notify_host(CHAN_TENNANT_HOSTS, host, acfg);
}


int list_subscriptions(const char *tenant,
		       const struct agent_config *acfg)
{
	char *filter;
	struct tbl *table;
	int r;
	const char *real_tenant = tenant ?: acfg->db.user;

	if (!api->get_table)
		return -EOPNOTSUPP;

	filter = strjoina("tennant = '", real_tenant, "'", NULL);

	table = api->get_table(TABLE_TENNANT_HOSTS, "*", filter, acfg);
	if (!table)
		LOGRAW("\nNo hosts subscribed to tennant %s\n", real_tenant);
	else {
		LOGRAW("\nHosts subscribed to tennant %s:\n", real_tenant);
		for (r = 0; r < table->rows; r ++)	
			LOGRAW("\t%s\n", table->value[r][0]);
	}

	free_tbl(table);
	return 0;
	
}

extern int change_host_state(const char *host, const char *newstate,
			     const struct agent_config *acfg)
{
	char *sql;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("UPDATE nodes SET state = '", newstate,
		       "' WHERE hostname = '", host, "'");

	return api->send_raw_sql(sql, acfg);
}

struct tbl* get_tennants_for_host(const char *host,
			  	  const struct agent_config *acfg)
{
	char *filter;

	if (!api->get_table)
		return NULL;

	filter = strjoina("hostname = '", host, "'", NULL);

	return api->get_table(TABLE_TENNANT_HOSTS, "*", filter, acfg);
}

struct tbl* get_repos_for_tennant(const char *tenant,
				  const struct agent_config *acfg)
{
	char *filter;

	if (!api->get_table)
		return NULL;

	filter = strjoina("tennant='", tenant, "'", NULL);

	return api->get_table(TABLE_YUM_CONFIG, "name, url", filter, acfg);
}

struct tbl* get_containers_for_host(const char *host,
				   const char *state,
				   const struct agent_config *acfg)
{
	char *filter;

	if (!api->get_table)
		return NULL;

	filter = strjoina("(hostname='", host, "' AND state='", state, "')", NULL);

	return api->get_table(TABLE_CONTAINERS, "*", filter, acfg);
}

static struct tbl *get_container_info(const char *iname,
				      const char *tennant,
                                     const struct agent_config *acfg)
{
	char *filter;

	if (!api->get_table)
		return NULL;

	filter = strjoina("(iname='", iname, "' AND tennant='", tennant,"')", NULL);

	return api->get_table(TABLE_CONTAINERS, "*", filter, acfg);
}

struct tbl *get_containers_of_type(const char *cname,
					  const char *tennant,
					  const char *host,
					  const struct agent_config *acfg)
{
	char *bfilter, *filter;

	if (!api->get_table)
		return NULL;

	bfilter = strjoina("(cname='", cname, "' AND tennant='", tennant, NULL);

	if (host)
		filter = strjoina(bfilter, "' AND hostname='", host, "')", NULL);
	else 
		filter = strjoina(bfilter, "')", NULL);

	return api->get_table(TABLE_CONTAINERS, "*", filter, acfg);
}

int request_create_container(const char *cname,
                             const char *iname,
                             const char *chost,
			     const char *tennant,
			     const struct agent_config *acfg)
{
	char *sql;
	int rc;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("INSERT INTO containers VALUES(",
		       "'", tennant, "',",
		       "'", iname, "',",
		       "'", cname, "',",
		       "'", chost, "',",
		       "'staged')", NULL);

	rc = api->send_raw_sql(sql, acfg);

	if (rc)
		return rc;

	/*
	 * If we add the sql safely, then we need to wake someone up to read the table
	 * If a chost is specified, then notify that host only, otherwise, notify the 
	 * master to pick a host for us
	 */
	if (chost)
		if (!strcmp(chost, "all"))
			rc = notify_tennant(CHAN_CONTAINERS, tennant, acfg);
		else
			rc = notify_host(CHAN_CONTAINERS, chost, acfg);
	else
		LOG(INFO, "NEED TO IMPLEMENT MASTER FUNCTIONALITY HERE\n");

	return rc;

}

int request_delete_container(const char *iname,
			     const char *tennant,
			     const int force,
			     const struct agent_config *acfg)
{
	char *sql;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("DELETE FROM containers WHERE tennant='",
		tennant, "' AND iname='",iname,
		"' AND (state='failed' OR state='staged')", NULL);

	return api->send_raw_sql(sql, acfg);

}

extern int request_boot_container(const char *iname, const char *tennant,
				  const struct agent_config *acfg)
{
	int rc;
	struct tbl *container;
	char *host;

	container = get_container_info(iname, tennant, acfg);

	if (is_tbl_empty(container)) {
		LOG(WARNING, "Container %s does not exist\n", iname);
		rc = -EEXIST;
		goto out;
	}

	rc = change_container_state(tennant, iname, "staged",
	 			    "start-requested", acfg);

	if (rc)
		rc = change_container_state(tennant, iname, "failed",
				    "start-requested", acfg);

	if (rc)
		LOG(WARNING, "container %s is in the wrong state\n", iname);

		
	host = lookup_tbl(container, 0, COL_HOSTNAME);

	if (host) {
		if(!strcmp(host, "all"))
			rc = notify_tennant(CHAN_CONTAINERS, tennant, acfg);
		else
			rc = notify_host(CHAN_CONTAINERS, host, acfg);
	} else {
		LOG(INFO, "NEED TO IMPLEMENT MASTER FUNCTION FOR HOST SELECTION\n");
	}

	if (rc)
		LOG(WARNING, "Notificaion of host failed.  Boot may be delayed\n");

out:
	free_tbl(container);
	return rc; 
	
}

extern int request_poweroff_container(const char *iname,
				      const char *tennant,
				  const struct agent_config *acfg)
{
	int rc;
	struct tbl *container;

	rc = change_container_state(tennant, iname, "running",
				    "exiting", acfg);

	container = get_container_info(iname, tennant, acfg);

	rc = notify_host(CHAN_CONTAINERS, lookup_tbl(container, 0, COL_HOSTNAME), acfg);

	return rc;
}

int print_container_list(const char *tennant, 
			 const struct agent_config *acfg)
{
	struct tbl *containers;
	char *filter;
	int i;

	if (!api->get_table)
		return -EOPNOTSUPP;

	filter = strjoina("tennant='",tennant,"'",NULL);

	containers = get_raw_table(TABLE_CONTAINERS, filter, acfg);

	if (!containers || !containers->rows)
		return -ENOENT;

	LOGRAW("CONTAINER NAME   |   CONTAINER TYPE   |   STATE\n");
	LOGRAW("--------------------------------------------\n");
	for (i=0; i < containers->rows; i++) {
		LOGRAW("%-17s|%-20s|%-8s\n",
			(char *)lookup_tbl(containers, i, COL_INAME),
			(char *)lookup_tbl(containers, i, COL_CNAME),
			(char *)lookup_tbl(containers, i, COL_STATE));
	}

	free_tbl(containers);
	return 0;

}



extern int change_container_state(const char *tennant,
                                  const char *iname,
				  const char *oldstate,
                                  const char *newstate,
                                  const struct agent_config *acfg)
{
	char *sql;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("UPDATE containers set state ='", newstate,
		       "' WHERE tennant = '", tennant,
		       "' AND iname = '", iname,
		       "' AND state = '", oldstate, "'", NULL);

	return api->send_raw_sql(sql, acfg);
}

extern int change_container_state_batch(const char *tennant,
					const char *oldstate,
					const char *newstate,
					const struct agent_config *acfg)
{
	char *sql;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("UPDATE containers set state ='", newstate,
		       "' WHERE tennant = '", tennant,
		       "' AND state = '", oldstate,
		       "' AND hostname = '", acfg->cmdline.hostname,
		       "'");

	return api->send_raw_sql(sql, acfg);
}

int delete_container(const char *iname, const char *tennant,
                     const struct agent_config *acfg)
{
	char *sql;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("DELETE FROM containers where tennant='", tennant,
		"' AND iname='", iname, "'", NULL);

	return api->send_raw_sql(sql, acfg);
}


int notify_host(const enum listen_channel chn, const char *host,
		const struct agent_config *acfg)
{
	char *name;

	if (!api->notify)
		return -EOPNOTSUPP;

	name = strjoina(channel_map[chn], "-", host, NULL);

	return api->notify(NOTIFY_HOST, chn, name, acfg);

} 

int notify_tennant(const enum listen_channel chn, const char *tennant,
		const struct agent_config *acfg)
{
	char *name;

	if (!api->notify)
		return -EOPNOTSUPP;

	name = strjoina(channel_map[chn], "-", tennant, NULL);

	return api->notify(NOTIFY_TENNANT, chn, name, acfg);
}

struct tbl* get_raw_table(enum db_table table, char *filter, const struct agent_config *acfg)
{
	if (!api->get_table)
		return NULL;

        return api->get_table(table, "*", filter, acfg);
}

int network_create_config(const char *name, const char *cfstring, const char *tennant, const struct agent_config *acfg)
{
	char *sql;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	/*
	 * until we implement network start/stop functions, we just mark
	 * all networks as active
	 */
	sql = strjoina("INSERT INTO networks VALUES('", name, "' , '"
			, tennant, "' , 'active', '", cfstring, "')", NULL);


	return api->send_raw_sql(sql, acfg);
	
}


int network_create(const char *name, const char *configfile, const char *tennant, const struct agent_config *acfg)
{
	struct stat buf;
	char *configbuf;
	FILE *configp;
	int rc = -ENOENT;

	/*
	 * Start by reading in the config file to a string
	 */

	if (stat(configfile, &buf)) {
		/* Problem with the config file */
		return errno;
	}

	if (buf.st_size > 8192) {
		/* hm, suspciuosly large config */
		return -ERANGE;
	}

	configbuf = calloc(1, buf.st_size);

	if (!configbuf)
		return -ENOMEM;

	configp = fopen(configfile, "r");

	if (!configp)
		goto out_free;

	rc = ERANGE;
	if (fread(configbuf, buf.st_size, 1, configp) != 1)
		goto out_close;

	rc = network_create_config(name, configbuf, tennant, acfg);

out_close:
	fclose(configp);
out_free:
	free(configbuf);
	return rc;

}

int network_delete(const char *name, const char *tennant, const struct agent_config *acfg)
{
	char *sql;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("DELETE FROM networks WHERE name='",name,"' AND tennant='",tennant,"'",NULL);

	return api->send_raw_sql(sql, acfg);
}

int network_list(const char *tennant, const struct agent_config *acfg)
{
	struct tbl *networks;
	char *filter;
	int i;

	filter = strjoina("tennant='",tennant,"'",NULL);

	LOGRAW("Available networks\n");
	
	networks = get_raw_table(TABLE_NETWORKS, filter, acfg);

	if (!networks)
		return 0;

	LOGRAW("NETWORK NAME   |   STATE\n");
	LOGRAW("--------------------------------------------\n");
	for (i=0; i < networks->rows; i++) {
		LOGRAW("%-17s|%-8s\n",
			(char *)lookup_tbl(networks, i, COL_NAME),
			(char *)lookup_tbl(networks, i, COL_STATE));
	}


	free_tbl(networks);
	return 0;	
}

int network_attach(const char *container, const char *network, const char *tennant, const struct agent_config *acfg)
{
	char *sql;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("INSERT INTO net_container_map VALUES('", tennant, "','",
		       container, "','", network, "')", NULL);

	return api->send_raw_sql(sql, acfg);
}


int network_detach(const char *container, const char *network, const char *tennant, const struct agent_config *acfg)
{
	char *sql;

	if (!api->send_raw_sql)
		return -EOPNOTSUPP;

	sql = strjoina("DELETE FROM net_container_map WHERE ",
		       "tennant='", tennant,"'",
		       "AND name='", container, "'",
		       "AND network='", network,"'", NULL);

	return api->send_raw_sql(sql, acfg);
}

struct tbl * get_network_info(const char *network, const char *tennant, const struct agent_config *acfg)
{
	struct tbl *networks;
	char *filter;

	filter = strjoina("tennant='",tennant,"' AND name='", network, "'",NULL);

	
	networks = get_raw_table(TABLE_NETWORKS, filter, acfg);

	if (!networks)
		return 0;

	if (networks->rows == 0) {
		free_tbl(networks);
		networks = NULL;
	}

	return networks;
}
